# -*- coding: utf-8 -*-
"""

# ======================================================================
# begin of file *Script_MPAPPtools_MPAPP.py*
# ======================================================================
# with functions from the publication Primpke, S., C. Lorenz, R. Rascher-Friesenhausen, and G. Gerdts. 2017. 'An automated approach for microplastics analysis using focal plane array (FPA) FTIR microscopy and image analysis', Anal. Methods, 9: 1499-511.
#sprimpke@awi.de (c) 2019

@author: sprimpke@awi.de
"""
# module imports
# =============================================================================

# standard library imports
import os.path                     # for path operations

# external third party imports
import numpy as np                 # for scientific computing
import pandas as pd                # for data anlysis
import SimpleITK as sitk           # for image analysis
import math as mt

# local imports
from tools_MPAPP_fastercrosshair import closing5x5pixel
from tools_MPAPP_faster          import closing4x4pixel
from tools_MPAPP_fastercrosshair import data2image
from tools_MPAPP_fastercrosshair import image2data
from tools_MPAPP_fastercrosshair import pdmv3x3pixel
from tools_MPAPP_fastercrosshair import calculate_distance
from tools_MPAPP_fastercrosshair import binary_invert
from tools_MPAPP_fastercrosshair import close_filaments
from tools_MPAPP_fastercrosshair import applymasktodata
from tools_MPAPP_fastercrosshair import image2data_setp

# constants, global vars
# =============================================================================
 # Generation of dataframe for size analysis
_C = ['single'] + [str(i) for i in range(25, 525, 25)] + ['large']
_C_10 = ['single'] + ['10'] + [str(i) for i in range(25, 525, 25)] + ['large']
_S = len(_C)*[0]
_S_10 = len(_C_10)*[0]
        
EMPTY = pd.DataFrame(_S, index=_C, columns=['s'])
EMPTY_10 = pd.DataFrame(_S_10, index=_C_10, columns=['s'])

EMPTY = pd.DataFrame(_S, index=_C, columns=['s'])
EMPTY_PARAMETER = pd.DataFrame(columns=['fileID','c','p','Material','Elongation','Diameter(MP)/Length(MF) in micron','Minimum diameter in micron','Volume im micron³','Weight']) 

_X = [str(i) for i in range(25, 5025, 25)] + ['large']
_Y = len(_X)*[0]
        
EMPTYFIBER = pd.DataFrame(_Y, index=_X, columns=['s'])

BULK_SET = set(list(range(1, 31)))
OTHER_SET = {15,20,25,28,30}
BIO_SET = {12,13,24,27}
POLYMER_SET = BULK_SET - OTHER_SET - BIO_SET


# class defs
# =============================================================================

class polymerimage_fromfile:
    def __init__(self, file_name, dir_name, pixelsize):
       
        # Polymer class dependent storage (polymer, bio or other)
        # This type of code is dependent on the database!
        # Database design is the reason for all the if cases
        
        self.base_name = get_base_name(dir_name)+'/'+get_base_name(file_name)
        
        # Important global constants     
        self.original_data = pd.read_csv(file_name, delimiter=';')
        
        self.fiberdatacombined = pd.DataFrame(columns=['x','y','g','p'])
        
        self.particledatacombined = pd.DataFrame(columns=['x','y','g','p'])
        
        self.skeletondatacombined = pd.DataFrame(columns=['x','y','g','p'])
        
        self.pixelsize = pixelsize
        
        self.counter=0
        
        name="threshold_and_metadata.csv"
        metadata= pd.read_csv(name, delimiter=';')
        
        self.img_size = max(self.original_data.x) + 1, max(self.original_data.y) + 1
        
        self.polymerlist = [polymerdata(p=i, pixelsize=self.pixelsize, data=self.original_data[self.original_data.p == i], savepath=self.base_name, img_size=self.img_size, density=metadata.Density[i], polymertypename = metadata.Polymername[i]) for i in range (0,31,1)]
        
        
        
        
    def analyse_global_data(self, save=False):
        """Analyse global data.

        Create polymer histogram and gray value image of global dataframe.

        data: data frame with columns 'x', 'y', 'g' and 'p'
        size: tuple of width and height
        name: name of data

        Return None.
        """
        # Write data in csv file
        if save:
            self.original_data.to_csv(self.base_name+"_image.csv", index=False, sep=';')

        # Create gray value image and write in png file
        img = data2image(self.original_data, size=self.img_size)
        write_image(img, self.base_name+"_all.png")

        # Determine polymer histogram and write in csv file
        p = self.original_data.p.value_counts().sort_index()
        p.to_csv(self.base_name+"_histogram.csv", sep=';')
        return    
    
    #def determinepolymers(self, test):
        #pass
    
    def analyse_polymers(self):
        if self.pixelsize > 10 :
            bulk = EMPTY.copy()
            polymer, bio, other = EMPTY.copy(), EMPTY.copy(), EMPTY.copy()

            #Particle analyses 
            hist_dict = {}
            for p in range(1, len(self.polymerlist)):
                if int(self.polymerlist[p].get_length()) > 0:  # if data is not empty
                    hist_c = self.polymerlist[p].analyse_local_data(output=True)
                    hist_dict[p] = hist_c
                    # Read out of Sizeclasses
                    sb = self.polymerlist[p].size_analysis(save=True)
                    # Counting for different datasets
                    bulk += sb
                    if p in POLYMER_SET:
                        polymer += sb
                    if p in OTHER_SET:
                        other += sb
                    if p in BIO_SET:
                        bio += sb
                else:
                    pass
        elif self.pixelsize < 10 :
            bulk = EMPTY_10.copy()
            polymer, bio, other = EMPTY_10.copy(), EMPTY_10.copy(), EMPTY_10.copy()

            #Particle analyses 
            hist_dict = {}
            for p in range(1, len(self.polymerlist)):
                if int(self.polymerlist[p].get_length()) > 0:  # if data is not empty
                    hist_c = self.polymerlist[p].analyse_local_data(output=True)
                    hist_dict[p] = hist_c
                    # Read out of Sizeclasses
                    sb = self.polymerlist[p].size_analysis(save=True)
                    # Counting for different datasets
                    bulk += sb
                    if p in POLYMER_SET:
                        polymer += sb
                    if p in OTHER_SET:
                        other += sb
                    if p in BIO_SET:
                        bio += sb
                else:
                    pass


        bulk.to_csv(self.base_name+"_sizeclasses_bulk.csv", sep=';')
        polymer.to_csv(self.base_name+"_sizeclasses_polymer.csv", sep=';')
        bio.to_csv(self.base_name+"_sizeclasses_bio.csv", sep=';')
        other.to_csv(self.base_name+"_sizeclasses_other.csv", sep=';')        # for later studies
        return

    def analyse_fibers(self):
        #Fibers analyses
        overall=pd.DataFrame(columns=['c','length'])
        for p in range(1, len(self.polymerlist)):
            if int(self.polymerlist[p].get_length()) > 0:  # if data is not empty
                #Starting fiber analysis
                overall=pd.concat([overall,self.polymerlist[p].predict_fibers()])              
            else:
                pass
        #Data readout
        overall.to_csv(self.base_name+"_fibers_complete.csv", sep=';')
        particledata1=pd.DataFrame(columns=['x','y','g','p'])
        fiberdata1=pd.DataFrame(columns=['x','y','g','p'])
        skeletondata1=pd.DataFrame(columns=['x','y','g','p'])
        for p in range(1, len(self.polymerlist)):
            particledata1=pd.concat([particledata1,self.polymerlist[p].get_particledata()])
            fiberdata1=pd.concat([fiberdata1,self.polymerlist[p].get_fiberdata()])
            skeletondata1=pd.concat([skeletondata1,self.polymerlist[p].get_skeletondata()])
        #Dataoutput for images
        fiberdata1.reset_index(drop=True,inplace=True)
        self.fiberdatacombined=fiberdata1
        self.fiberdatacombined.to_csv(self.base_name+"_fibers_raw_bulk.csv",index=False, sep=';')
        particledata1.reset_index(drop=True,inplace=True)
        self.particledatacombined=particledata1
        self.particledatacombined.to_csv(self.base_name+"_particles_bulk.csv",index=False, sep=';')
        skeletondata1.reset_index(drop=True,inplace=True)
        self.skeletondatacombined=skeletondata1
        self.skeletondatacombined.to_csv(self.base_name+"_fibers_skeleton_bulk.csv",index=False, sep=';')
        return

    
    def get_filename(self):
        name=self.base_name
        return name
    
    def get_data(self):
        data=self.original_data
        return data
    
    def generate_countfile(self):
        statscontainer={'p':[''],'stat':[self.base_name],'mass':['g']}
        statistics=pd.DataFrame(statscontainer, columns=['p','stat','mass'])  
        for i in range(1, len(self.polymerlist)):
            statscontainer2={'p':[i],'stat':[int(self.polymerlist[i].get_count())],'mass':[float(self.polymerlist[i].get_mass())]}
            statistics2=pd.DataFrame(statscontainer2, columns=['p','stat','mass'])
            statistics=pd.concat([statistics,statistics2])
        statistics.to_csv(self.base_name+"_particles_and_mass_perpolymer.csv", sep=';')
    
    def generate_fulloutput(self):
        #Generation of one excelfile with all particle data 
        overall=self.polymerlist[1].get_sizeclasses()
        parameters_overall = self.polymerlist[1].get_sizeparamterlist()
        for i in range(2, len(self.polymerlist)):
            overall=pd.concat([overall,self.polymerlist[i].get_sizeclasses()],axis=1)
            parameters_overall=pd.concat([parameters_overall,self.polymerlist[i].get_sizeparamterlist()])
        overall.to_csv(self.base_name+"_sizeclasses_forcalculations.csv", sep=';')
        parameters_overall.to_csv(self.base_name+"_particle_parameters.csv", sep=';')
        
    def generate_fullfiberoutput(self):
        #Generation of one excelfile with all particle data        
        overall=self.polymerlist[1].get_fibersizeclasses()
        for i in range(2, len(self.polymerlist)):
            overall=pd.concat([overall,self.polymerlist[i].get_fibersizeclasses()],axis=1)
        overall.to_csv(self.base_name+"_fiber_sizeclasses_forcalculations.csv", sep=';')
    
    def get_fiberdatacombined(self):
        data =self.fiberdatacombined
        return data
    
    def get_particledatacombined(self):
        data =self.particledatacombined
        return data
        
class polymerimage_fromdata(polymerimage_fromfile):
    def __init__(self, data, filepath, pixelsize):
       
        # Polymer class dependent storage (polymer, bio or other)
        # This type of code is dependent on the database!
        # Database design is the reason for all the if cases
        
        self.base_name = filepath
        # Important global constants     
        self.original_data = data
        
        self.pixelsize = pixelsize
        
        name="threshold_and_metadata.csv"
        metadata= pd.read_csv(name, delimiter=';')
        
        
        self.counter=0
        
        self.img_size = max(self.original_data.x) + 1, max(self.original_data.y) + 1
        
        self.polymerlist = [polymerdata(p=i, pixelsize=self.pixelsize, data=self.original_data[self.original_data.p == i], savepath=self.base_name, img_size=self.img_size, density=metadata.Density[i], polymertypename = metadata.Polymername[i]) for i in range (0,31,1)]

class polymerimage_fromMPhunterfile(polymerimage_fromfile):
    def __init__(self, file_name, dir_name, pixelsize):
       
        # Polymer class dependent storage (polymer, bio or other)
        # This type of code is dependent on the database!
        # Database design is the reason for all the if cases
        
        self.base_name = get_base_name(dir_name)+'/'+get_base_name(file_name)
        
        df = pd.read_csv(file_name, delimiter=';')
        
        df_container={'x':['0'],'y':['0'],'g':['0'],'p':['0'],'ID':['0']}
        data=pd.DataFrame(df_container, columns=['x','y','g','p','ID'])
        j=0
        for i in df.index:    
            if int(df.p1[i]) == int(df.p2[i]):
                if float(df.g1[i]) > 0.3 and float(df.g2[i]) > 0.3:
                    g_sum=int(float(df.g1[i])*1000+float(df.g2[i])*1000)
                    data.loc[j]=[df.x[i],df.y[i],g_sum,df.p2[i],df.DatabaseID2[i]]
                    j=j+1
        save_name = "%s_forqc.csv" % (self.base_name)
        data.to_csv(save_name,sep=';', index=False)
        # Important global constants     
        self.original_data = data
        
        self.fiberdatacombined = pd.DataFrame(columns=['x','y','g','p'])
        
        self.particledatacombined = pd.DataFrame(columns=['x','y','g','p'])
        
        self.skeletondatacombined = pd.DataFrame(columns=['x','y','g','p'])
        
        self.pixelsize = pixelsize
        
        self.counter=0
        
        self.img_size = max(self.original_data.x) + 1, max(self.original_data.y) + 1
        
        name="threshold_and_metadata.csv"
        metadata= pd.read_csv(name, delimiter=';')
        
        
        self.polymerlist = [polymerdata(p=i, pixelsize=self.pixelsize, data=self.original_data[self.original_data.p == i], savepath=self.base_name, img_size=self.img_size, density=metadata.Density[i], polymertypename = metadata.Polymername[i]) for i in range (0,31,1)]
        
                
        
# Definition of the class polymerdata
# This object allows the polymer specific data storage before and after closing for further analysis
# independent from external dataframes    
class polymerdata:
    def __init__(self, p, pixelsize,  data, savepath, img_size,density, polymertypename,saveimage = True):
        self.id = p
        self.pixelsize = pixelsize
        self.rawdata = data
        self.fiberdata = pd.DataFrame(columns=['x','y','g','p'])
        self.particledata = pd.DataFrame(columns=['x','y','g','p'])
        self.fiber_skeleton = pd.DataFrame(columns=['x','y','g','p'])
        self.count = 0
        self.datapath = savepath
        self.density=density
        self.mass = 0
        self.polymertypename=polymertypename
        self.analyzedsizeclass = EMPTY.copy()
        self.analyzedfibersizes = EMPTYFIBER.copy()
        self.parameters = EMPTY_PARAMETER.copy()
        self.length=len(self.rawdata.x)
        if self.length > 0:
            self.img_size = img_size
            self.rawimage = data2image(data, self.img_size)
            if saveimage:
                save_name = "%s_%02d.png" % (self.datapath, self.id)
                print(save_name)
                write_image(self.rawimage, save_name)
        
    def size_analysis(self, save=True):
        # Size class analysis
        # Internal counter will be reseted
        self.count=0
        if self.pixelsize > 10:
            sb = EMPTY.copy()
            for sidx in self.hist_c.index:
                area = float(self.pixelsize)*float(self.pixelsize)*sidx
                # for 1 pixel
                j = 1
                counted = False
                sizeclass = float(self.pixelsize)*float(self.pixelsize)
                if area == sizeclass:
                    sb.s[j-1] = self.hist_c[sidx]
                    counted = True
                    self.count += self.hist_c[sidx]
                j = j + 1
                # for values higher 1 pixel
                while j < 22 and not counted:
                    sizeclass = (25*(j-1))*(25*(j-1))
                    if area <= sizeclass:
                        sb.s[j-1] += self.hist_c[sidx]
                        counted = True
                        self.count += self.hist_c[sidx]
                    j = j + 1
                # if still higher it will be targeted in an extra field
                if j == 22 and not counted:
                    sb.s[j-1] += self.hist_c[sidx]
                    self.count += self.hist_c[sidx]
                    counted = True
        elif self.pixelsize > 1:
            sb = EMPTY_10.copy()
            for sidx in self.hist_c.index:
                area = float(self.pixelsize)*float(self.pixelsize)*sidx
                # for 1 pixel
                j = 1
                counted = False
                sizeclass = float(self.pixelsize)*float(self.pixelsize)
                if area == sizeclass:
                    sb.s[j-1] = self.hist_c[sidx]
                    counted = True
                    self.count += self.hist_c[sidx]
                j = j + 1
                if area <= 100:
                    sb.s[j-1] += self.hist_c[sidx]
                    counted = True
                    self.count += self.hist_c[sidx] 
                j= j + 1
                # for values higher 1 pixel
                while j < 23 and not counted:
                    sizeclass = (25*(j-2))*(25*(j-2))
                    if area <= sizeclass:
                        sb.s[j-1] += self.hist_c[sidx]
                        counted = True
                        self.count += self.hist_c[sidx]
                    j = j + 1
                # if still higher it will be targeted in an extra field
                if j == 23 and not counted:
                    sb.s[j-1] += self.hist_c[sidx]
                    self.count += self.hist_c[sidx]
                    counted = True            
            
        # Still to introduce Blank correction!
        # Data output
        if save:
            save_name = "%s_sizeclasses%02d.csv" % (self.datapath, self.id)
            sb.to_csv(save_name, sep=';')
        self.analyzedsizeclass = sb
        return sb
    
    def fiber_length_analysis(self, data):
        img=data2image(data,self.imgsize)
        cimg = sitk.ConnectedComponent(img, True)
        rimg = sitk.RelabelComponent(cimg)
    
        s1 = pd.Series(np.zeros(len(data)), index=data.index, name='c')     
        data_c = pd.concat([data, s1], axis=1)
    
        for i in data_c.index:
            x, y = int(data_c.x[i]), int(data_c.y[i])
            data_c.at[i, 'c'] = rimg[x, y]
    
        counts = data_c.c.value_counts()
        counts_df=pd.DataFrame(counts, columns=['c'])
        counts_df.reset_index(inplace=True)
        counts_df.rename(index=str,columns={"index":"c","c":"pixel"},inplace=True)
       
        for i in counts_df.index:
            pixel=counts_df.pixel[i]
            c_copy=counts_df.c[i]
            rawsize=self.pixelsize
            counts_df.loc[i]=[c_copy,round(pixel*rawsize)]
        
       
        
    def analyse_local_data(self, output=True):
        """Analyse particle data.
    
        Create particle gray value image, connected components histogram and
        data file.
    
        data:
        size:
        name:
        p:
        
        Return component histogram.
        """
    
        # Find connected components with 8 pixel neighboorhood
        cimg = sitk.ConnectedComponent(self.rawimage, True)
        rimg = sitk.RelabelComponent(cimg)
    
        # Append component column to data
        c = pd.Series(np.zeros(len(self.rawdata)), index=self.rawdata.index, name='c') # temporary
        data_c = pd.concat([self.rawdata, c], axis=1)
        for i in data_c.index:
            x, y = int(data_c.x[i]), int(data_c.y[i])
            data_c.at[i, 'c'] = rimg[x, y]
        if output:
            save_name = "%s_image%02d.csv" % (self.datapath, self.id)
            data_c.to_csv(save_name, sep=';',index=False)

        beta = sitk.LabelImageToLabelMap(rimg) #Generation of Labelmap
        beta3 = sitk.LabelMapToLabel(beta) # and transfer to Label data
        beta4 = sitk.LabelShapeStatisticsImageFilter()
        beta4.SetBackgroundValue(0)
        beta4.ComputeFeretDiameterOn()
        beta4.Execute(beta3) #Setup for Shape analysis of individual PFO
        counts = data_c.c.value_counts()
        
        labeldata=EMPTY_PARAMETER.copy() #Output for PFO data
        if output:
            for idx in range (1, beta4.GetNumberOfLabels()+1):
                el = beta4.GetElongation(label=idx)
                if el > 0 and el < 100000 and counts[idx] > 2:
                    length = beta4.GetFeretDiameter(label=idx)*self.pixelsize
                    ml=(length/el)
                    vl=(4/3)*mt.pi*(length/2)*(ml/2)*((ml*0.7)/2)
                    weight=(self.density/(10000*10000*10000))*vl
                    self.mass += weight
                    labeldata.loc[idx]=['self.fileID',idx,self.id,self.polymertypename,el,length,ml,vl,weight] #Readout of all data for Output
                else:
                    length=counts[idx]*self.pixelsize
                    ml=(self.pixelsize)
                    vl=(4/3)*mt.pi*(length/2)*(ml/2)*((ml*0.7)/2)
                    weight=(self.density/(10000*10000*10000))*vl
                    self.mass += weight
                    labeldata.loc[idx]=['self.fileID',idx,self.id,self.polymertypename,counts[idx],length,ml,vl,weight] #Readout of all data
            #save_name = "%s_particle_parameters%02d.csv" % (self.datapath, self.id)
            self.parameters = labeldata
            #labeldata.to_csv(save_name, sep=';', index = False)  
        # Determine particle components histogram and write in csv filehgh

        # Determine particle components histogram and write in csv file
        #counts = data_c.c.value_counts()
        counts_df=pd.DataFrame(counts, columns=['c'])
        counts_df.reset_index(inplace=True)
        counts_df.rename(index=str,columns={"index":"c","c":"pixel"},inplace=True)
        for i in counts_df.index:
            pixel=counts_df.pixel[i]
            c_copy=counts_df.c[i]
            rawsize=self.pixelsize*self.pixelsize
            counts_df.loc[i]=[c_copy,round(pixel*rawsize)]
        if output:
            save_name = "%s_particle_indvidual%02d.csv" % (self.datapath, self.id)
            counts_df.to_csv(save_name, sep=';', index = False)
        hist_c = counts.value_counts().sort_index()
        if output:
            save_name = "%s_particle_histogram%02d.csv" % (self.datapath, self.id)
            hist_c.to_csv(save_name, sep=';')
        self.hist_c = hist_c
        return hist_c

    def predict_fibers(self):
        counts_df=pd.DataFrame(columns={"c","length"}) # Preparation of DataFrame for PixelThresholdFilter
        size_img = sitk.ConnectedComponent(self.rawimage, True) # Size determination of particles
        rimg = sitk.RelabelComponent(size_img) # Relabeling for data analysis 
    
        # Append component column to data
        c = pd.Series(np.zeros(len(self.rawdata)), index=self.rawdata.index, name='c') # temporary
        data_c = pd.concat([self.rawdata, c], axis=1)
        for i in data_c.index:
            x, y = int(data_c.x[i]), int(data_c.y[i])
            data_c.at[i, 'c'] = rimg[x, y]
        counts = data_c.c.value_counts() #Transfer of the data for PixelThreshold
        
        
        data_fiber = self.rawdata[self.rawdata.p>256] # Generation of x,y,g,p data header for empty data frame
        for i in counts.index: #PixelThresholdFilter
            if counts[i] > 2: # MFT parametrt
                data_c_fiber=data_c[data_c.c == i]
                data_fiber=pd.concat([data_fiber,data_c_fiber]) #Transfer of potential fiber objects (PFO) into new data frame
        if len(data_fiber) != 0: # If PFO present fruther analysis is started
            img = data2image(data_fiber,self.img_size) #Generation of a new image only cointaining PFO
            bimg = sitk.BinaryThresholdImageFilter() #Preparation of Binarization
            aimg = bimg.Execute(img) # Binarization of the image
            
            aimg=binary_invert(aimg) #Closing of holes within the objects, Inversion necessary due to nature of the FPA images
            filling=sitk.BinaryFillholeImageFilter()
            cimg=filling.Execute(aimg)
            cimg=binary_invert(cimg)
            
            dimg=calculate_distance(cimg) # Distance Map Caluclation
            
            thres = sitk.BinaryThresholdImageFilter()
            thres.SetLowerThreshold(0)
            thres.SetUpperThreshold(4) # Distance Factor DF
            eimg = thres.Execute(dimg) #DistanceMask
            #save_name = "%s_%02d_development3.png" % (self.datapath, self.id)
            #print(save_name)
            #write_image(eimg, save_name) #Export of the image for beta testing
            
            for i in range(0,4,1): # Dilation with range until DL factor
                dilute=sitk.BinaryErodeImageFilter() # Erode was used instead of a binary inversion before and after the dilution process 
                eimg=dilute.Execute(eimg)
            #save_name = "%s_%02d_development4.png" % (self.datapath, self.id)
            #print(save_name)
            #write_image(eimg, save_name)
            mask=sitk.MaskImageFilter()
            fimg=mask.Execute(img,eimg) #Removing Particles via DistanceMap Mask
            
            beta_img = sitk.ConnectedComponent(fimg, True) #Begin data preparation for ElongationThesholdFilter
            dev_img = sitk.RelabelComponent(beta_img)
            beta = sitk.LabelImageToLabelMap(dev_img) #Generation of Labelmap
            beta3 = sitk.LabelMapToLabel(beta) # and transfer to Label data
            beta4 = sitk.LabelShapeStatisticsImageFilter()
            beta4.SetBackgroundValue(0)
            beta4.ComputeFeretDiameterOn()
            beta4.Execute(beta3) #Setup for Shape analysis of individual PFO
            beta3_data=image2data(dev_img) # Data transfer for access on DataFrame level
            c = pd.Series(np.zeros(len(beta3_data)), index=beta3_data.index, name='c') # temporary
            beta3_data_c = pd.concat([beta3_data, c], axis=1)
            for i in beta3_data_c.index:
                x, y = int(beta3_data_c.x[i]), int(beta3_data_c.y[i])
                beta3_data_c.at[i, 'c'] = dev_img[x, y] #Data preparation
            
            data_nonparticle = pd.DataFrame(columns=['x','y','g','c']) #Generation of new dataframe
            labeldata=pd.DataFrame(columns=['c','Elongation','FeretDiameter','Roundness']) #Output for PFO data 
            for idx in range (1, beta4.GetNumberOfLabels()+1):
                labeldata.loc[idx]=[idx,beta4.GetElongation(label=idx),beta4.GetFeretDiameter(label=idx),beta4.GetRoundness(label=idx)] #Readout of all data for Output

                if beta4.GetElongation(label=idx) >=3 and beta4.GetRoundness(label=idx) <=1: #ElongationThesholdFilter with ETP of 3
                    data_c_fiber=beta3_data_c[beta3_data_c.c == idx]
                    data_nonparticle=pd.concat([data_nonparticle,data_c_fiber]) #Data transfer into new Dataframe
            
            for idx in data_nonparticle.index:
                data_nonparticle.at[idx,'g']=1000 #Necessary for BinaryThresholdFilter, otherwise a bug is caused

            if len(data_nonparticle) != 0:
                
                non_particle=data2image(data_nonparticle,self.img_size) #Datatranfer fromDataframe into Image
                lefiber = sitk.BinaryThresholdImageFilter()

                binary_non_particle=lefiber.Execute(non_particle)

                binary_non_particle=binary_invert(binary_non_particle) #Preparation of Mask Containing to remove all nonPFO from the image 
                
                gimg=mask.Execute(fimg,binary_non_particle) #Application of the Mask
                pre_img=bimg.Execute(gimg) #Preparation of Data for Skeletonize Process

                pre_gimg=binary_invert(pre_img)
                skelton=sitk.BinaryThinningImageFilter() #Skeletonize Process
                skeltonized=skelton.Execute(pre_gimg)

            
                filaments_closed = close_filaments(skeltonized) #Clsoing of Filament Gaps
                save_name = "%s_%02d_fibers_skeleton_closed.png" % (self.datapath, self.id)
                print(save_name)
                write_image(filaments_closed, save_name)#Export of the image for beta testing
                
                self.fiber_skeleton=image2data_setp(filaments_closed,self.id) #Storage of Data
                
                
                
                fiberdata=image2data(filaments_closed) #Size analysis
                fiber_img = sitk.ConnectedComponent(filaments_closed, True)
                rimg = sitk.RelabelComponent(fiber_img)
            
                c = pd.Series(np.zeros(len(fiberdata)), index=fiberdata.index, name='c') # temporary
                data_c = pd.concat([fiberdata, c], axis=1)
                for i in data_c.index:
                    x, y = int(data_c.x[i]), int(data_c.y[i])
                    data_c.at[i, 'c'] = rimg[x, y]     
                    # Determine particle components histogram and write in csv file
                counts = data_c.c.value_counts()
                counts_df=pd.DataFrame(counts, columns=['c'])
                counts_df.reset_index(inplace=True)
                counts_df.rename(index=str,columns={"index":"c","c":"length"},inplace=True) #Readout for Transfer into .csv files
                for i in counts_df.index:
                    pixel=counts_df.length[i]
                    c_copy=counts_df.c[i]
                    rawsize=self.pixelsize
                    counts_df.loc[i]=[c_copy,round(pixel*rawsize)] 
                #save_name = "%s_fiber_indvidual%02d.csv" % (self.datapath, self.id) #Generation of .csv files
                #counts_df.to_csv(save_name, sep=';', index = False)
                hist_c = counts.value_counts().sort_index() #Size hsitrogramm for each found pixelsize
                #save_name = "%s_fiber_histogram%02d.csv" % (self.datapath, self.id)
                #hist_c.to_csv(save_name, sep=';')
                
                labeldata=EMPTY_PARAMETER.copy() #Output for PFO data
                for idx in counts_df.index:
                    length = counts_df.length[idx]
                    vl=mt.pi*((7.5)*(7.5))*(length) #an anverage diameter  of 30 miron is assumend 
                    weight=(self.density/(10000*10000*10000))*vl
                    self.mass += weight
                    #el=0
                    labeldata.loc[idx]=[self.fileID,idx,self.id,self.polymertypename,'NA','NA',length,15,vl,weight] #Readout of all data for Output

                #save_name = "%s_particle_parameters%02d.csv" % (self.datapath, self.id)
                self.parametersfibers = labeldata
                
                sb = EMPTYFIBER.copy() # Generation of Fiber sizeclasses
                for sidx in hist_c.index:
                    
                    length = float(self.pixelsize)*sidx
                    # for 1 pixel
                    counted = False
                    j=1
                    if length > 5000:
                        sb.s[200] = hist_c[sidx]
                        counted = True
                        # for values higher 1 pixel
                    while length <= 5000 and not counted:
                        sizeclass = j *25
                        if length <= sizeclass:
                            sb.s[j-1] += hist_c[sidx]
                            counted = True
                        j = j + 1
                # Data output
                
                self.analyzedfibersizes = sb
                
                save_name = "%s_sizeclasses_fiber%02d.csv" % (self.datapath, self.id)
                sb.to_csv(save_name, sep=';')
                
                save_name = "%s_%02d_fibers_raw.png" % (self.datapath, self.id)
                self.fiberdata=applymasktodata(data_fiber,binary_non_particle,self.img_size,save_name)
            
                save_name = "%s_%02d_particles.png" % (self.datapath, self.id)
                particlemask=binary_invert(pre_gimg)
                self.particledata=applymasktodata(self.rawdata,particlemask,self.img_size,save_name) #Determination of Particle Data
            
            else:
                save_name = "%s_%02d_particles.png" % (self.datapath, self.id) # Data transfer for particles if no PFO beyond ElongationThresholdFilter
                self.particledata=self.rawdata
                write_image(self.rawimage,save_name)
        else:
            save_name = "%s_%02d_particles.png" % (self.datapath, self.id) # Data transfer for particles if no PFO beyond ParticleThresholdFilter
            self.particledata=self.rawdata
            write_image(self.rawimage,save_name)
            
        return counts_df
       
    def get_sizeclasses(self):
        size = self.analyzedsizeclass
        return size
    
    def get_fibersizeclasses(self):
        size = self.analyzedfibersizes
        return size
        
    def get_count(self):
        count = self.count
        return count
        
    def get_rawdata(self):
        data =self.rawdata
        return data
        
    def get_length(self):
        length =self.length
        return length
    
    def get_fiberdata(self):
        data =self.fiberdata
        return data
    
    def get_particledata(self):
        data =self.particledata
        return data
    
    def get_skeletondata(self):
        data = self.fiber_skeleton
        return data
    
    def get_mass(self):
        mass =self.mass
        return mass
    
    def get_sizeparamterlist(self):
        parameters =self.parameters
        return parameters
    
# fcn def
# =============================================================================    

def write_image(img, name):
    Img = sitk.RescaleIntensity(img) #, outputMinimum=0, outputMaximum=255)
    Img = sitk.Cast(Img, sitk.sitkUInt8)
    #Img = sitk.InvertIntensity(Img)
    sitk.WriteImage(Img, name)
    return

def get_base_name(file_name):
    """Split file name and return base name part.

    file_name: a name of a file in the system

    base_name: base name of the file without dir and ext
    """
    dir_name, base_name = os.path.split(file_name)
    base_name, ext = os.path.splitext(base_name)
    return base_name
    
def apply_threshold(data):
    threshold_name="threshold_and_metadata.csv"
    threshold = pd.read_csv(threshold_name, delimiter=';')
    data_threshold = data[data.p>32] # Erstellen der x,y,g,p Zeile ohne Daten
    for i in range(0, max(data.p)+1):
        data_p_i=data[data.p ==i]
        data_p_threshold=data_p_i[data_p_i.g > threshold.t[i]]
        data_threshold=pd.concat([data_threshold,data_p_threshold])
    return data_threshold

def apply_thresholdFTIR(data):
    threshold_name="threshold_and_metadata_FTIR.csv"
    threshold = pd.read_csv(threshold_name, delimiter=';')
    data_threshold = data[data.p>32] # Erstellen der x,y,g,p Zeile ohne Daten
    for i in range(0, max(data.p)+1):
        data_p_i=data[data.p ==i]
        data_p_threshold=data_p_i[data_p_i.g > threshold.t[i]]
        data_threshold=pd.concat([data_threshold,data_p_threshold])
    return data_threshold    

def analyse(file_name, dir_name, ext_pixelsize):
    """Data analysis"""

    # get base name of data file
    # Published part in OOP

    # -------------------------------------------------------------------------
    print("- Analyse original data")
    # set save name head

    print("  -- Analyse particles")

    pre_data = polymerimage_fromfile(file_name,dir_name,ext_pixelsize)
    
    threshold = apply_threshold(pre_data.get_data())
    
        #AAP data Output (Additional output supressed by commenting the functions, for use just remove # starting from raw data)
    
    rawdata = polymerimage_fromdata(threshold ,pre_data.get_filename()+'_raw',ext_pixelsize)
    
    rawdata.analyse_global_data()
    rawdata.analyse_polymers()
    rawdata.generate_countfile()
    rawdata.generate_fulloutput()
    
        #MPAPP data Output
    #rawdata.analyse_fibers()
    
    #rawfiberimage=closedimage.get_fiberdatacombined()
    #rawparticleimage=rawdata.get_particledatacombined()
    
    #raw_particle = polymerimage_fromdata(rawparticleimage,pre_data.get_filename()+'_raw_particles',ext_pixelsize)
    
    #raw_particle.analyse_global_data()
    #raw_particle.analyse_polymers()
    #raw_particle.generate_countfile()
    #raw_particle.generate_fulloutput()   
    
    print("- Closing Particles")
    #closed_particles = closing4x4pixel(threshold, 5) # with 5 iterations
    closed_particles = closing5x5pixel(threshold, 5) # with 5 iterations
    print("- ... done")
    
    closedimage = polymerimage_fromdata(closed_particles,pre_data.get_filename()+'_closed',ext_pixelsize)
        #AAP data Output
    closedimage.analyse_global_data(save=True)
    closedimage.analyse_polymers()
    closedimage.generate_countfile()
    closedimage.generate_fulloutput()
     
        #MPAPP data Output
    #closedimage.analyse_fibers()
    
    #particleimage=closedimage.get_particledatacombined()
    
    #particle = polymerimage_fromdata(particleimage,pre_data.get_filename()+'_closed_particles',ext_pixelsize)
    

    
    #particle.analyse_global_data()
    #particle.analyse_polymers()
    #particle.generate_countfile()
    #particle.generate_fulloutput()
    

    print("- Median data")
    #pdmv_data=pdmv3x3pixel(closed_particles, 5)
        # An approach to find single pixels within a area of different polymer assigments while keeping single pixels
        #AAP data Output
    print("- ... done")
    #pdmv_image= polymerimage_fromdata(pdmv_data,pre_data.get_filename()+'_median',ext_pixelsize)
    #median_image.analyse_global_data(save=True)
    #median_image.analyse_polymers()
    #median_image.generate_countfile()
    #median_image.generate_fulloutput() 
     #MPAPP data Output
    #pdmv_image.analyse_fibers()
    
    #rawfiberimage=closedimage.get_fiberdatacombined()
    #pdmv_image_particleimage=pdmv_image.get_particledatacombined()
    
    #pdmv_image_particle = polymerimage_fromdata(pdmv_image_particleimage,pre_data.get_filename()+'_median(final)_particles',ext_pixelsize)
    
    #pdmv_image_particle.analyse_global_data()
   # pdmv_image_particle.analyse_polymers()
    #pdmv_image_particle.generate_countfile()
   # pdmv_image_particle.generate_fulloutput()
   # pdmv_image.generate_fullfiberoutput()
    


    return

def analyse_dataframe(data, filename, ext_pixelsize):
    """Data analysis"""

    # get base name of data file
    # Published part in OOP

    # -------------------------------------------------------------------------
    print("- Analyse original data")
    # set save name head

    print("  -- Analyse particles")

    pre_data = polymerimage_fromdata(data,filename,ext_pixelsize)
    
    threshold = apply_threshold(pre_data.get_data())
    
        #AAP data Output (Additional output supressed by commenting the functions, for use just remove # starting from raw data)
    
    rawdata = polymerimage_fromdata(threshold ,pre_data.get_filename()+'_raw',ext_pixelsize)
    
    rawdata.analyse_global_data()
    rawdata.analyse_polymers()
    rawdata.generate_countfile()
    rawdata.generate_fulloutput()
    
        #MPAPP data Output
    #rawdata.analyse_fibers()
    
    #rawfiberimage=closedimage.get_fiberdatacombined()
    #rawparticleimage=rawdata.get_particledatacombined()
    
    #raw_particle = polymerimage_fromdata(rawparticleimage,pre_data.get_filename()+'_raw_particles',ext_pixelsize)
    
    #raw_particle.analyse_global_data()
    #raw_particle.analyse_polymers()
    #raw_particle.generate_countfile()
    #raw_particle.generate_fulloutput()   
    
    print("- Closing Particles")
    #closed_particles = closing4x4pixel(threshold, 5) # with 5 iterations
    closed_particles = closing5x5pixel(threshold, 5) # with 5 iterations
    print("- ... done")
    
    closedimage = polymerimage_fromdata(closed_particles,pre_data.get_filename()+'_closed',ext_pixelsize)
        #AAP data Output
    closedimage.analyse_global_data(save=True)
    closedimage.analyse_polymers()
    closedimage.generate_countfile()
    closedimage.generate_fulloutput()
     
        #MPAPP data Output
    #closedimage.analyse_fibers()
    
    #particleimage=closedimage.get_particledatacombined()
    
    #particle = polymerimage_fromdata(particleimage,pre_data.get_filename()+'_closed_particles',ext_pixelsize)
    

    
    #particle.analyse_global_data()
    #particle.analyse_polymers()
    #particle.generate_countfile()
    #particle.generate_fulloutput()
    

    print("- Median data")
    #pdmv_data=pdmv3x3pixel(closed_particles, 5)
        # An approach to find single pixels within a area of different polymer assigments while keeping single pixels
        #AAP data Output
    print("- ... done")
    #pdmv_image= polymerimage_fromdata(pdmv_data,pre_data.get_filename()+'_median',ext_pixelsize)
    #median_image.analyse_global_data(save=True)
    #median_image.analyse_polymers()
    #median_image.generate_countfile()
    #median_image.generate_fulloutput() 
     #MPAPP data Output
    #pdmv_image.analyse_fibers()
    
    #rawfiberimage=closedimage.get_fiberdatacombined()
    #pdmv_image_particleimage=pdmv_image.get_particledatacombined()
    
    #pdmv_image_particle = polymerimage_fromdata(pdmv_image_particleimage,pre_data.get_filename()+'_median(final)_particles',ext_pixelsize)
    
    #pdmv_image_particle.analyse_global_data()
    #pdmv_image_particle.analyse_polymers()
    #pdmv_image_particle.generate_countfile()
    #pdmv_image_particle.generate_fulloutput()
    #pdmv_image.generate_fullfiberoutput()
    
    return
